@@ -8,6 +8,7 @@ https://github.com/chapmanb/bcbio-nextgen/blob/master/bcbio/distributed/ipython.
 """
 import contextlib
 import copy
+import math
 import os
 import pipes
 import uuid
@@ -16,7 +17,6 @@ import subprocess
 import time
 from distutils.version import LooseVersion
 import sys
-import imp
 
 from IPython.parallel import Client
 from IPython.parallel.apps import launcher
@@ -132,6 +132,18 @@ controller_cmd_argv = [sys.executable, "-E", "-c"] + \
                       ["; ".join(resource_cmds + ["from cluster_helper.cluster import VMFixIPControllerApp",
                                                   "VMFixIPControllerApp.launch_instance()"])]
 
+def get_engine_commands(context, n):
+    """Retrieve potentially multiple engines running in a single submit script.
+
+    Need to background the initial processes if multiple run.
+    """
+    assert n > 0
+    engine_cmd_full = '%s %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"' % \
+                      (' '.join(map(pipes.quote, engine_cmd_argv)), ' '.join(timeout_params))
+    out = [engine_cmd_full.format(**context)]
+    for _ in range(n - 1):
+        out.insert(0, "(%s &) &&" % (engine_cmd_full.format(**context)))
+    return "\n".join(out)
 
 # ## Platform LSF
 class BcbioLSFEngineSetLauncher(launcher.LSFEngineSetLauncher):
@@ -139,6 +151,7 @@ class BcbioLSFEngineSetLauncher(launcher.LSFEngineSetLauncher):
     """
     batch_file_name = Unicode(unicode("lsf_engine" + str(uuid.uuid4())))
     cores = traitlets.Integer(1, config=True)
+    numengines = traitlets.Integer(1, config=True)
     mem = traitlets.Unicode("", config=True)
     tag = traitlets.Unicode("", config=True)
     resources = traitlets.Unicode("", config=True)
@@ -152,12 +165,11 @@ class BcbioLSFEngineSetLauncher(launcher.LSFEngineSetLauncher):
 #BSUB -R "span[hosts=1]"
 {mem}
 {resources}
-%s %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
-    """ % (' '.join(map(pipes.quote, engine_cmd_argv)),
-           ' '.join(timeout_params)))
+{cmd}
+""")
 
     def start(self, n):
-        self.context["cores"] = self.cores
+        self.context["cores"] = self.cores * self.numengines
         if self.mem:
             # lsf.conf can specify nonstandard units for memory reservation
             lsf_unit = lsf.get_lsf_units(resource=True)
@@ -165,11 +177,13 @@ class BcbioLSFEngineSetLauncher(launcher.LSFEngineSetLauncher):
             # check if memory reservation is per core or per job
             if lsf.per_core_reservation():
                 mem = mem / self.cores
+            mem = mem * self.numengines
             self.context["mem"] = '#BSUB -R "rusage[mem=%s]"' % mem
         else:
             self.context["mem"] = ""
         self.context["tag"] = self.tag if self.tag else "bcbio"
         self.context["resources"] = _format_lsf_resources(self.resources)
+        self.context["cmd"] = get_engine_commands(self.context, self.numengines)
         return super(BcbioLSFEngineSetLauncher, self).start(n)
 
 def _format_lsf_resources(resources):
@@ -189,18 +203,21 @@ def _format_lsf_resources(resources):
 class BcbioLSFControllerLauncher(launcher.LSFControllerLauncher):
     batch_file_name = Unicode(unicode("lsf_controller" + str(uuid.uuid4())))
     tag = traitlets.Unicode("", config=True)
+    cores = traitlets.Integer(1, config=True)
     resources = traitlets.Unicode("", config=True)
     job_array_template = Unicode('')
     queue_template = Unicode('')
     default_template = traitlets.Unicode("""#!/bin/sh
 #BSUB -q {queue}
 #BSUB -J {tag}-c
+#BSUB -n {cores}
 #BSUB -oo bcbio-ipcontroller.bsub.%%J
 {resources}
 %s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" %s
     """ % (' '.join(map(pipes.quote, controller_cmd_argv)),
            ' '.join(controller_params)))
     def start(self):
+        self.context["cores"] = self.cores
         self.context["tag"] = self.tag if self.tag else "bcbio"
         self.context["resources"] = _format_lsf_resources(self.resources)
         return super(BcbioLSFControllerLauncher, self).start()
@@ -272,6 +289,8 @@ echo \($SGE_TASK_ID - 1\) \* 0.5 | bc | xargs sleep
 class BcbioSGEControllerLauncher(launcher.SGEControllerLauncher):
     batch_file_name = Unicode(unicode("sge_controller" + str(uuid.uuid4())))
     tag = traitlets.Unicode("", config=True)
+    cores = traitlets.Integer(1, config=True)
+    pename = traitlets.Unicode("", config=True)
     resources = traitlets.Unicode("", config=True)
     queue_template = Unicode('')
     default_template = traitlets.Unicode(u"""#$ -V
@@ -279,6 +298,7 @@ class BcbioSGEControllerLauncher(launcher.SGEControllerLauncher):
 #$ -w w
 #$ -S /bin/sh
 #$ -N {tag}-c
+{cores}
 {queue}
 {resources}
 {exports}
@@ -286,6 +306,7 @@ class BcbioSGEControllerLauncher(launcher.SGEControllerLauncher):
 """ % (' '.join(map(pipes.quote, controller_cmd_argv)),
        ' '.join(controller_params)))
     def start(self):
+        self.context["cores"] = "#$ -pe %s %s" % (self.pename, self.cores) if self.cores > 1 else ""
         self.context["tag"] = self.tag if self.tag else "bcbio"
         self.context["resources"] = "\n".join([_prep_sge_resource(r)
                                                for r in str(self.resources).split(";")
@@ -415,6 +436,7 @@ class BcbioSLURMEngineSetLauncher(SLURMLauncher, launcher.BatchClusterAppMixin):
     batch_file_name = Unicode(unicode("SLURM_engine" + str(uuid.uuid4())))
     machines = traitlets.Integer(0, config=True)
     cores = traitlets.Integer(1, config=True)
+    numengines = traitlets.Integer(1, config=True)
     mem = traitlets.Unicode("", config=True)
     tag = traitlets.Unicode("", config=True)
     account = traitlets.Unicode("", config=True)
@@ -432,16 +454,15 @@ class BcbioSLURMEngineSetLauncher(SLURMLauncher, launcher.BatchClusterAppMixin):
 {machines}
 {mem}
 {resources}
-%s %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
-    """ % (' '.join(map(pipes.quote, engine_cmd_argv)),
-           ' '.join(timeout_params)))
+{cmd}
+""")
 
     def start(self, n):
-        self.context["cores"] = self.cores
+        self.context["cores"] = self.cores * self.numengines
         if self.mem:
-            self.context["mem"] = "#SBATCH --mem=%s" % int(float(self.mem) * 1024.0)
+            self.context["mem"] = "#SBATCH --mem=%s" % int(float(self.mem) * 1024.0 * self.numengines)
         else:
-            self.context["mem"] = "#SBATCH --mem=%d" % int(DEFAULT_MEM_PER_CPU * self.cores)
+            self.context["mem"] = "#SBATCH --mem=%d" % int(DEFAULT_MEM_PER_CPU * self.cores * self.numengines)
         self.context["tag"] = self.tag if self.tag else "bcbio"
         self.context["machines"] = ("#SBATCH %s" % (self.machines) if int(self.machines) > 0 else "")
         self.context["account"] = self.account
@@ -449,11 +470,13 @@ class BcbioSLURMEngineSetLauncher(SLURMLauncher, launcher.BatchClusterAppMixin):
         self.context["resources"] = "\n".join(["#SBATCH --%s" % r.strip()
                                                for r in str(self.resources).split(";")
                                                if r.strip()])
+        self.context["cmd"] = get_engine_commands(self.context, self.numengines)
         return super(BcbioSLURMEngineSetLauncher, self).start(n)
 
 class BcbioSLURMControllerLauncher(SLURMLauncher, launcher.BatchClusterAppMixin):
     batch_file_name = Unicode(unicode("SLURM_controller" + str(uuid.uuid4())))
     account = traitlets.Unicode("", config=True)
+    cores = traitlets.Integer(1, config=True)
     timelimit = traitlets.Unicode("", config=True)
     mem = traitlets.Unicode("", config=True)
     tag = traitlets.Unicode("", config=True)
@@ -464,6 +487,7 @@ class BcbioSLURMControllerLauncher(SLURMLauncher, launcher.BatchClusterAppMixin)
 #SBATCH -e bcbio-ipcontroller.err.%%j
 #SBATCH -A {account}
 #SBATCH -t {timelimit}
+#SBATCH --cpus-per-task={cores}
 {mem}
 {resources}
 %s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" %s
@@ -472,6 +496,7 @@ class BcbioSLURMControllerLauncher(SLURMLauncher, launcher.BatchClusterAppMixin)
     def start(self):
         self.context["account"] = self.account
         self.context["timelimit"] = self.timelimit
+        self.context["cores"] = self.cores
         self.context["mem"] = "#SBATCH --mem=%d" % (8 * DEFAULT_MEM_PER_CPU)
         self.context["tag"] = self.tag if self.tag else "bcbio"
         self.context["resources"] = "\n".join(["#SBATCH --%s" % r.strip()
@@ -526,53 +551,6 @@ class BcbioOLDSLURMControllerLauncher(SLURMLauncher, launcher.BatchClusterAppMix
         self.context["timelimit"] = self.timelimit
         return super(BcbioOLDSLURMControllerLauncher, self).start(1)
 
-
-# ## PBS
-class BcbioPBSEngineSetLauncher(launcher.PBSEngineSetLauncher):
-    """Custom launcher handling heterogeneous clusters on PBS.
-    """
-    batch_file_name = Unicode(unicode("pbs_engines" + str(uuid.uuid4())))
-    cores = traitlets.Integer(1, config=True)
-    mem = traitlets.Unicode("", config=True)
-    tag = traitlets.Unicode("", config=True)
-    pename = traitlets.Unicode("", config=True)
-    resources = traitlets.Unicode("", config=True)
-    default_template = traitlets.Unicode("""#PBS -V
-#PBS -j oe
-#PBS -S /bin/sh
-#PBS -q {queue}
-#PBS -N {tag}-e
-#PBS -t 1-{n}
-{mem}
-%s %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
-""" % (' '.join(map(pipes.quote, engine_cmd_argv)),
-       ' '.join(timeout_params)))
-
-    def start(self, n):
-        self.context["cores"] = self.cores
-        self.context["pename"] = str(self.pename)
-        if self.mem:
-            self.context["mem"] = "#PBS -l mem=%smb" % int(float(self.mem) * 1024)
-        else:
-            self.context["mem"] = ""
-        self.context["tag"] = self.tag if self.tag else "bcbio"
-        return super(BcbioPBSEngineSetLauncher, self).start(n)
-
-
-class BcbioPBSControllerLauncher(launcher.PBSControllerLauncher):
-    batch_file_name = Unicode(unicode("pbs_controller" + str(uuid.uuid4())))
-    tag = traitlets.Unicode("", config=True)
-    default_template = traitlets.Unicode(u"""#PBS -V
-#PBS -S /bin/sh
-#PBS -N {tag}-c
-%s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" %s
-""" % (' '.join(map(pipes.quote, controller_cmd_argv)),
-       ' '.join(controller_params)))
-
-    def start(self):
-        self.context["tag"] = self.tag if self.tag else "bcbio"
-        return super(BcbioPBSControllerLauncher, self).start()
-
 # ## Torque
 class TORQUELauncher(launcher.BatchSystemLauncher):
     """A BatchSystemLauncher subclass for Torque"""
@@ -651,12 +629,14 @@ class BcbioTORQUEControllerLauncher(TORQUELauncher, launcher.BatchClusterAppMixi
     """Launch a controller using Torque."""
     batch_file_name = Unicode(unicode("torque_controller" + str(uuid.uuid4())),
                               config=True, help="batch file name for the engine(s) job.")
+    cores = traitlets.Integer(1, config=True)
     tag = traitlets.Unicode("", config=True)
     resources = traitlets.Unicode("", config=True)
     default_template = Unicode("""#!/bin/sh
 #PBS -V
 #PBS -N {tag}-c
 #PBS -j oe
+#PBS -l nodes=1:ppn={cores}
 {resources}
 cd $PBS_O_WORKDIR
 %s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" %s
@@ -666,6 +646,7 @@ cd $PBS_O_WORKDIR
     def start(self):
         """Start the controller by profile or profile_dir."""
         try:
+            self.context["cores"] = self.cores
             self.context["tag"] = self.tag if self.tag else "bcbio"
             self.context["resources"] = "\n".join(_prep_torque_resources(self.resources))
             return super(BcbioTORQUEControllerLauncher, self).start(1)
@@ -684,9 +665,12 @@ class BcbioPBSPROEngineSetLauncher(PBSPROLauncher, launcher.BatchClusterAppMixin
     batch_file_name = Unicode(u'pbspro_engines', config=True,
                               help="batch file name for the engine(s) job.")
     tag = traitlets.Unicode("", config=True)
+    cores = traitlets.Integer(1, config=True)
+    mem = traitlets.Unicode("", config=True)
     default_template = Unicode(u"""#!/bin/sh
 #PBS -V
 #PBS -N {tag}-e
+{cores}
 cd $PBS_O_WORKDIR
 %s %s --profile-dir="{profile_dir}" --cluster-id="{cluster_id}"
     """ % (' '.join(map(pipes.quote, engine_cmd_argv)),
@@ -694,6 +678,8 @@ cd $PBS_O_WORKDIR
 
     def start(self, n):
         """Start n engines by profile or profile_dir."""
+        self.context["cores"] = ("#PBS -l select=1:ncpus=%s;mem=%smb" % (self.cores, self.mem)
+                                 if self.cores > 1 and self.mem else "")
         self.context["tag"] = self.tag if self.tag else "bcbio"
         return super(BcbioPBSPROEngineSetLauncher, self).start(n)
 
@@ -704,9 +690,11 @@ class BcbioPBSPROControllerLauncher(PBSPROLauncher, launcher.BatchClusterAppMixi
     batch_file_name = Unicode(u'pbspro_controller', config=True,
                               help="batch file name for the controller job.")
     tag = traitlets.Unicode("", config=True)
+    cores = traitlets.Integer(1, config=True)
     default_template = Unicode("""#!/bin/sh
 #PBS -V
 #PBS -N {tag}-c
+{cores}
 cd $PBS_O_WORKDIR
 %s --ip=* --log-to-file --profile-dir="{profile_dir}" --cluster-id="{cluster_id}" %s
 """ % (' '.join(map(pipes.quote, controller_cmd_argv)),
@@ -714,6 +702,7 @@ cd $PBS_O_WORKDIR
 
     def start(self):
         """Start the controller by profile or profile_dir."""
+        self.context["cores"] = "#PBS -l select=1:ncpus=%s" % self.cores if self.cores > 1 else ""
         self.context["tag"] = self.tag if self.tag else "bcbio"
         return super(BcbioPBSPROControllerLauncher, self).start(1)
 
@@ -728,13 +717,22 @@ def _scheduler_resources(scheduler, params, queue):
     """Retrieve custom resource tweaks for specific schedulers.
     Handles SGE parallel environments, which allow multicore jobs
     but are specific to different environments.
+    Pulls out hacks to work in different environments:
+      - mincores -- Require a minimum number of cores when submitting jobs
+                    to avoid single core jobs on constrained queues
     """
-    resources = copy.deepcopy(params.get("resources", []))
+    orig_resources = copy.deepcopy(params.get("resources", []))
     specials = {}
-    if not resources:
-        resources = []
-    if isinstance(resources, basestring):
-        resources = resources.split(";")
+    if not orig_resources:
+        orig_resources = []
+    if isinstance(orig_resources, basestring):
+        orig_resources = orig_resources.split(";")
+    resources = []
+    for r in orig_resources:
+        if r.startswith("mincores="):
+            specials["mincores"] = int(r.split("=")[-1])
+        else:
+            resources.append(r)
     if scheduler in ["SGE"]:
         pass_resources = []
         for r in resources:
@@ -773,6 +771,10 @@ def _start(scheduler, profile, queue, num_jobs, cores_per_job, cluster_id,
         resources, slurm_atrs = get_slurm_attributes(queue, resources)
     else:
         slurm_atrs = None
+    mincores = specials.get("mincores", 1)
+    if mincores > cores_per_job:
+        mincores = int(math.ceil(mincores / float(cores_per_job)))
+        num_jobs = int(math.ceil(num_jobs / float(mincores)))
 
     args = cluster_cmd_argv + \
         ["start",
@@ -782,6 +784,7 @@ def _start(scheduler, profile, queue, num_jobs, cores_per_job, cluster_id,
          "--log-to-file",
          "--debug",
          "--n=%s" % num_jobs,
+         "--%s.cores=%s" % (controller_class, min(mincores, 2)),
          "--%s.cores=%s" % (engine_class, cores_per_job),
          "--%s.resources='%s'" % (controller_class, resources),
          "--%s.resources='%s'" % (engine_class, resources),
@@ -795,7 +798,10 @@ def _start(scheduler, profile, queue, num_jobs, cores_per_job, cluster_id,
          "--cluster-id=%s" % (cluster_id)
          ]
     args += _get_profile_args(profile)
+    if mincores > 1 and mincores > cores_per_job:
+        args += ["--%s.numengines=%s" % (engine_class, mincores)]
     if specials.get("pename"):
+        args += ["--%s.pename=%s" % (controller_class, specials["pename"])]
         args += ["--%s.pename=%s" % (engine_class, specials["pename"])]
     if specials.get("memtype"):
         args += ["--%s.memtype=%s" % (engine_class, specials["memtype"])]
@@ -822,6 +828,9 @@ def _start_local(cores, profile, cluster_id):
     subprocess.check_call(args)
     return cluster_id
 
+def stop_from_view(view):
+    _stop(view.clusterhelper["profile"], view.clusterhelper["cluster_id"])
+
 def _stop(profile, cluster_id):
     args = cluster_cmd_argv + \
            ["stop", "--cluster-id=%s" % cluster_id]
@@ -899,6 +908,7 @@ def cluster_view(scheduler, queue, num_jobs, cores_per_job=1, profile=None,
             view = _get_direct_view(client, retries)
         else:
             view = _get_balanced_blocked_view(client, retries)
+        view.clusterhelper = {"profile": profile, "cluster_id": cluster_id}
         if dill:
             pickleutil.use_dill()
             view.apply(pickleutil.use_dill)
@@ -950,7 +960,11 @@ def _slurm_is_old():
 def _slurm_version():
     version_line = subprocess.Popen("sinfo -V", shell=True,
                                     stdout=subprocess.PIPE).communicate()[0]
-    return version_line.split()[1]
+    parts = version_line.split()
+    if len(parts) > 0:
+        return version_line.split()[1]
+    else:
+        return "2.6+"
 
 # ## Temporary profile management
 
